"stop_time": <%= @stop_time.to_i %>,
"elapsed": <%= @elapsed.to_i %>,
"requests_done": <%= @requests_done.to_i %>,
"cached_requests": <%= @cached_requests.to_i %>,
"response_status_codes": <%= @response_status_codes.to_json %>,
"response_status_codes_warning": <%= @response_status_codes_warning %>,
"response_status_codes_warnings": <%= @response_status_codes_warnings.to_json %>,
"data_sent": <%= @data_sent.to_i %>,
"data_sent_humanised": <%= @data_sent.bytes_to_human.to_json %>,
"data_received": <%= @data_received.to_i %>,
"data_received_humanised": <%= @data_received.bytes_to_human.to_json %>,
"used_memory": <%= @used_memory.to_i %>,
"used_memory_humanised": <%= @used_memory.bytes_to_human.to_json %>,