"backup_folders": {
<% unless @backup_folders.empty? -%>
<% last_index = @backup_folders.size - 1 -%>
<% @backup_folders.each_with_index do |backup_folder, index| -%>
  <%= backup_folder.url.to_json %>: {
    "severity": <%= backup_folder.severity.to_json %>,
    <%= render('@finding', item: backup_folder) -%>
  }<% unless index == last_index -%>,<% end -%>
<% end -%>
<% end -%>
},