 | <%= critical_icon %> Title: <%= @v.title %>
<% if @v.uuid -%>
 |     UUID: <%= @v.uuid %>
<% end -%>
<% if @v.cvss -%>
 |     CVSS: <%= @v.cvss[:score] %> (<%= @v.cvss[:vector] %>)
<% end -%>
<% if @v.fixed_in -%>
 |     Fixed in: <%= @v.fixed_in %>
<% end -%>
<% if @v.poc -%>
 |     PoC: available
<% end -%>
<% unless (references = @v.references_urls).empty? -%>
<% if references.size == 1 -%>
 |     Reference: <%= references.first %>
<% else -%>
 |     References:
<% references.each do |ref| -%>
 |      - <%= ref %>
<% end -%>
<% end -%>
<% end -%>