<%= info_icon %> Finished: <%= @stop_time.asctime %>
<%= info_icon %> Requests Done: <%= @requests_done %>
<%= info_icon %> Cached Requests: <%= @cached_requests %>
<% if @response_status_codes && !@response_status_codes.empty? -%>
<%= info_icon %> Most response codes received: <%= @response_status_codes.map { |label, count| "#{label}: #{count}" }.join(', ') %>
<% end -%>
<% if @response_status_codes_warnings && !@response_status_codes_warnings.empty? -%>
<% @response_status_codes_warnings.each do |warning| -%>
<%= warning_icon %> <%= warning %>
<% end -%>
<% end -%>
<%= info_icon %> Data Sent: <%= @data_sent.bytes_to_human %>
<%= info_icon %> Data Received: <%= @data_received.bytes_to_human %>
<%= info_icon %> Memory used: <%= @used_memory.bytes_to_human %>
<%= info_icon %> Elapsed time: <%= Time.at(@elapsed).utc.strftime('%H:%M:%S') %>
