# frozen_string_literal: true

# = Public Suffix
#
# Domain name parser based on the Public Suffix List.
#
# Copyright (c) 2009-2026 Simone Carletti <weppos@weppos.net>

module PublicSuffix

  # @return [String] the current library version
  VERSION = "7.0.5"

end
