"start_time": <%= @start_time.to_i %>,
"start_memory": <%= @start_memory.to_i %>,
"command_line": <%= "wpscan #{@command_line}".to_json %>,
"hostname": <%= @hostname.to_s.to_json %>,
"target_url": <%= @url.to_s.to_json %>,
"target_ip": <%= @ip.to_s.to_json %>,
"effective_url": <%= @effective_url.to_s.to_json %>,
<% if WPScan::ParsedCli.exclude_vulns && !WPScan::ParsedCli.exclude_vulns.empty? -%>
"excluded_vulnerabilities": <%= Array(WPScan::ParsedCli.exclude_vulns).to_json %>,
<% end -%>