 | Location: <%= @wp_item.url %>
<% if @wp_item.latest_version && !@wp_item.outdated? -%>
 | Latest Version: <%= @wp_item.latest_version %><% if @wp_item.version %> (up to date)<% end %>
<% end -%>
<% if @wp_item.last_updated -%>
 | Last Updated: <%= @wp_item.last_updated_display %><%= @wp_item.last_updated_cli_suffix %>
<% end -%>
<% if @wp_item.active_installs -%>
 | Active Installs: <%= ActiveSupport::NumberHelper.number_to_delimited(@wp_item.active_installs) %> (per WordPress.org)
<% end -%>
<% if @wp_item.readme_url -%>
 | Readme: <%= @wp_item.readme_url %>
<% end -%>
<% if @wp_item.latest_version && @wp_item.outdated? -%>
 | <%= warning_icon %> The version is out of date, the latest version is <%= @wp_item.latest_version %>
<% end -%>
<% if @wp_item.directory_listing? -%>
 | <%= critical_icon %> Directory listing is enabled
<% end -%>
<% if @wp_item.error_log? -%>
 | <%= critical_icon %> An error log file has been found: <%= @wp_item.url('error_log') %>
<% end -%>