"plugin": {
  <%= @plugin.slug.to_json %>: {
    <%= render('@wp_item', wp_item: @plugin) %>,
    <%= render('@finding', item: @plugin) -%>,
    <% if @plugin.version -%>
    "version": {
      "number": <%= @plugin.version.number.to_json %>,
      "confidence": <%= @plugin.version.confidence.to_json %>,
      <%= render('@finding', item: @plugin.version) -%>
    }
    <% else -%>
    "version": null
    <% end -%>
  }
}
