<%= bar_clear -%>

<%= info_icon %> <%= @plugin %>
<%= render('@wp_item', wp_item: @plugin) -%>
 |
<%= render('@finding', item: @plugin) -%>
 |
<% if @plugin.version -%>
 | Version: <%= @plugin.version %> (<%= @plugin.version.confidence %>% confidence)
<%= render('@finding', item: @plugin.version) -%>
<% else -%>
 | The version could not be determined.
<% end -%>
