# frozen_string_literal: true

require_relative "time/acts_like"
require_relative "time/calculations"
require_relative "time/compatibility"
require_relative "time/conversions"
require_relative "time/zones"
