<% if @backup_folders.empty? -%>
<%= notice_icon %> No Backup Folders Found.
<% else -%>
<%= notice_icon %> Backup Folder(s) Identified:
<% @backup_folders.each do |backup_folder| -%>

<% icon = backup_folder.severity == :high ? critical_icon : warning_icon -%>
<%= icon %> <%= backup_folder %>
<%= render('@finding', item: backup_folder) -%>
<% end -%>
<% end %>